import torch
import torchvision
import torchvision.transforms as transforms
import matplotlib.pyplot as plt
import numpy as np


transform = transforms.Compose(
    [transforms.ToTensor(),
     transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))])

trainset = torchvision.datasets.CIFAR10(root='./data', train=True,
                                        download=True, transform=transform)
trainloader = torch.utils.data.DataLoader(trainset, batch_size=100,
                                          shuffle=True, num_workers=2)

testset = torchvision.datasets.CIFAR10(root='./data', train=False,
                                       download=True, transform=transform)
testloader = torch.utils.data.DataLoader(testset, batch_size=10000,
                                         shuffle=False, num_workers=2)

classes = ('plane', 'car', 'bird', 'cat',
           'deer', 'dog', 'frog', 'horse', 'ship', 'truck')


def imshow(img):
    img = img / 2 + 0.5     # unnormalize
    npimg = img.numpy()
    plt.imshow(np.transpose(npimg, (1, 2, 0)))
    plt.show()

# get some random training images
dataiter = iter(trainloader)
images, labels = dataiter.next()
inputs = images.to(device)
print(device)
# show images
imshow(torchvision.utils.make_grid(images))
# print labels
print(' '.join('%5s' % classes[labels[j]] for j in range(4)))

test_images, test_labels = iter(testloader).next()

def test(images, labels):
    correct = 0
    total = 0
    with torch.no_grad():
        images = images.to(device)
        labels = labels.to(device)

        outputs = model(images, False)
        _, predicted = torch.max(outputs.data, 1)
        total += labels.size(0)
        correct += (predicted == labels).sum().item()

        # print('Accuracy of the network on the 10000 test images: %d %%' % (
        #     100 * correct / total))
    return correct / total


# from alex.alex.checkpoint import Checkpoint

# C = Checkpoint("examples/configs/small1.yml",
#                ["cache",  "config_1622420349826577.json"],
#                ["checkpoints", None])

# ckpt = C.load()

model = Model()

model.to(device)

optimizer = model.get_optimizer(model.params)

learning_rate = model.get_scheduler(optimizer)


for epoch in range(90):  # loop over the dataset multiple times

    running_loss = 0.0
    for i, data in enumerate(trainloader, 0):
        # get the inputs; data is a list of [inputs, labels]

        inputs, labels = data

        inputs = inputs.to(device)
        labels = labels.to(device)
        # zero the parameter gradients
        optimizer.zero_grad()

        # forward + backward + optimize
        output = model(inputs, True)
        loss = model.get_loss(model.trainable_params, [labels, output])
        loss.backward()
        optimizer.step()


        # print statistics
        running_loss += loss.item()
        if i % 500 == 499:    # print every 2000 mini-batches
            accuracy = test(test_images, test_labels)
            print('[%d, %5d] accuracy: %.3f, loss: %.3f' %
                  (epoch + 1, i + 1, accuracy, running_loss / 500))
            # C.save(model.trainable_params)
            running_loss = 0.0
    learning_rate.step()
print('Finished Training')



# ----------------------------------------------------------------------
# Created: sön apr 25 15:24:51 2021 (+0200)
# Last-Updated:
# Filename: test_schema.py
# Author: Yinan Yu
# Description:
# ----------------------------------------------------------------------
import unittest
from pprint import pprint
import os
from copy import deepcopy
from glob import glob
from jsonschema import validate

from alex.alex import (schema, util, const)


class TestSchema(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.conv_expected = {'filters':
                              {'shape':
                               {'kernel_size_w': 3,
                                'kernel_size_h': 3,
                                'input_shape': 'channels',
                                'n_filters': 64},
                               'initializer':
                               {'xavier_uniform': {'seed': None}}},
                              'padding': [1, 1],
                              'strides': [2, 2],
                              'use_bias': False,
                              'dilation': 1}


    def test_expand_ingredient(self):
        conv = util.read_yaml(util.get_component_path("conv"))
        conv_schema = util.read_yaml(util.get_component_spec_path("conv"))["definitions"]["conv"]
        validate(conv, conv_schema)
        conv_invalid = deepcopy(self.conv_expected)
        conv_invalid["filters"]["shape"]["input_shape"] = "not_valid_value"
        with self.assertRaises(Exception) as context:
            validate(conv_invalid, conv_schema)

        conv_invalid = deepcopy(self.conv_expected)
        conv_invalid["invalid_key"] = None
        with self.assertRaises(Exception) as context:
            validate(conv_invalid, conv_schema)


    def test_schemas(self):
        """A fuzzy test to check if the schema is reasonable.
           The assumption is that the component is well formed."""
        all_schemas = sorted(glob(os.path.join(const.COMPONENT_BASE_PATH, ".*.yml")))
        for _schema in all_schemas:
            component_name = _schema.split("/")[-1].split(".")[1]
            component = util.read_yaml(util.get_component_path(component_name))
            component_schema = util.read_yaml(util.get_component_spec_path(component_name))["definitions"][component_name]
            validate(component, component_schema)
            print(component_name, "schema checking")


if __name__ == '__main__':
    unittest.main()



# ----------------------------------------------------------------------
# Created: lör maj 15 18:31:09 2021 (+0200)
# Last-Updated:
# Filename: annotator_interface.py
# Author: Yinan Yu
# Description:
# ----------------------------------------------------------------------
from abc import ABC, abstractmethod
import collections
from pprint import pprint
from copy import deepcopy
from alex.alex import core, dsl_parser


class Annotator(ABC):

    def __init__(self, config_path, exclude_types=[], naive=True):
        if isinstance(config_path, str):
            self.config_path = config_path
            self.components_list = dsl_parser.parse(self.config_path)
        elif isinstance(config_path, list):
            self.components_list = config_path
        self.tree = core.alex_graph_to_tree(dsl_parser.list_to_graph(self.components_list),
                                            exclude_types=exclude_types,
                                            naive=naive)
        self.components = dsl_parser.list_to_dict(self.components_list)

        self.passes = [None]
        self.anno_name = "unnamed annotation"

    def get_input_names(self, name, tree):
        node = tree[name]
        if name == "root":
            return None
        elif node["type"] == "recipe":
            recipe_name = node["name"]
            first_ingredient_in_recipe = list(filter(lambda x: recipe_name in x,
                                                     self.components))[0]
            inputs = self.components[first_ingredient_in_recipe]["meta"]["inputs"]
        elif name in self.components:
            inputs = self.components[name]["meta"]["inputs"]
        else:
            parent = core.get_ancestor_ingredient_node(node, tree, "name")
            inputs = self.components[parent]["meta"]["inputs"]
        return inputs

    def get_input_nodes(self, name, tree):
        inputs = self.get_input_names(name, tree)
        if inputs is not None:
            inputs = core.get_nodes(tree, inputs)
        return inputs

    def annotate_tree(self):

        self.annotated = deepcopy(self.tree)

        for ipass, operation in enumerate(self.passes):
            if operation is None:
                raise Exception("Register an operation in each pass")
            if not isinstance(operation, list):
                operation = [operation]
            for sub_op in operation:
                for name in self.annotated:
                    node = self.annotated[name]
                    self.annotated[name] = sub_op(node)
        #     print("Annotation pass %i done" % ipass)

        print("Annotation %s done" % self.anno_name)

        return self.annotated



# ----------------------------------------------------------------------
# Created: ons maj 19 00:43:05 2021 (+0200)
# Last-Updated:
# Filename: complexity_measure.py
# Author: Yinan Yu
# Description:
# ----------------------------------------------------------------------
from abc import ABC, abstractmethod
from math import floor
from pprint import pprint
from copy import deepcopy
from alex.alex import core, const, node_interface
from alex.alex.annotator_interface import Annotator
from alex.engine import ns_alex
from math import floor
from pprint import pprint
from copy import deepcopy
from alex.engine import ns_alex
from jsonschema import validate
import collections
import numpy as np
import collections
import os
import numpy as np

# There are two main complexity measure we use here
# the Halstead complexity measure and the Cyclomatic measure
def cyclomatic(e, n, c):
    """
    e: the number of edges
    n: the number of nodes
    c: the connected components
    """
    return e - n + 2*c


def halstead(n_1, n_2, N_1, N_2):
    """
    n_1: the number of unique operators
    n_2: the number of unique operands
    N_1: the total number of operators
    N_2: the total number of operands

    """
    length = N_1 + N_2
    vocabulary = n_1 + n_2
    if N_1 == 0 or N_2 == 0:
        volume = 0
        difficulty = 0
    else:
        volume = round(length * np.log2(vocabulary), 2)
        difficulty = round(n_1/2 * N_2/n_2, 2)
    effort = difficulty * volume
    return {"vocabulary": vocabulary,
            "length": length,
            "volume": volume,
            "difficulty": difficulty,
            "effort": effort,
            "time": effort/18,
            "bugs": effort**(2/3)/3000}


def node_complextiy(node, tree, field):
    name = node["name"]
    descendants = core.get_descendants(name, tree)
    operators = list(map(lambda y: y["value"],
                         filter(lambda x: len(x["children"]) != 0,
                                descendants)))
    operands = list(map(lambda x: x["value"], descendants))
    n_1 = len(set(operators)) + 1
    n_2 = len(set(operands))
    N_1 = len(operators) + 1
    N_2 = len(operands)
    metric = halstead(n_1, n_2, N_1, N_2)
    if field is None:
        return metric
    else:
        return metric[field]


class Ingredient(node_interface.Ingredient):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def calculate_complexity(node, tree, field):
        """Halstead"""
        return node_complextiy(node, tree, field=field)


class Recipe(node_interface.Recipe):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def calculate_complexity(self, node, tree, *args, **kwargs):
        """Cyclomatic
        Default is sequential
        """
        if node["value"] != "root":
            n_nodes = self.get_node_count(node)
            n_edges = self.get_edge_count(node)
            # Assuming that all nodes are connected in a recipe
            metric = cyclomatic(n_edges, n_nodes, n_nodes)
        else:
            descendants = core.get_descendants(node["name"], tree, "complexity")
            metric = round(sum(descendants), 2)
        return metric


class Hyperparam(node_interface.Hyperparam):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def calculate_complexity(node, tree, field):
        """Halstead"""
        return node_complextiy(node, tree, field=field)


# ------------------------------------------------------------------------ #
def nodes(node):
    value = node["value"]
    _nodes = {"ingredient": Ingredient,
              "hyperparam": Hyperparam,
              "recipe": Recipe}
    if value not in _nodes:
        value = node["type"]
    return _nodes[value](node["value"])


class ComplexityMeasure(Annotator):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.anno_name = "complexity"
        self.passes = [[self.annotate]]

    def annotate(self, node):
        node = deepcopy(node)
        node[self.anno_name] = nodes(node).calculate_complexity(node,
                                                                self.annotated,
                                                                field="volume")
        return node



# Created: mån maj 17 11:46:49 2021 (+0200)
# ----------------------------------------------------------------------
# Last-Updated:
# Filename: node_interface.py
# Author: Yinan Yu
# Description:
# ----------------------------------------------------------------------
from abc import ABC, abstractmethod
from math import floor
import collections
import os
import numpy as np
from pprint import pprint
from copy import deepcopy
from alex.alex import core, const, util, dsl_parser
from alex.engine import ns_alex
from jsonschema import validate


# These classes are only caching static info
class Node:
    @staticmethod
    def get_children_names(node):
        return node["children"]


class Ingredient(Node):

    def __init__(self, ingredient_name):
        self.name = ingredient_name

    def annotate(self):
        pass


class Recipe(Node):
    def __init__(self, recipe_name):
        self.name = recipe_name
        if self.name != "root":
            self.config_path = os.path.join(const.COMPONENT_BASE_PATH,
                                            self.name + ".yml")
            self.recipe = dsl_parser.parse(self.config_path, return_dict=False)
        else:
            self.recipe = None

    def get_node_count(self, node=None):
        if self.recipe is None:
            return len(node["children"])
        else:
            return len(self.recipe)

    def get_edge_count(self, node=None):
        if self.recipe is not None:
            return sum(list(map(lambda x: len(x["meta"][const.INPUTS]),
                                self.recipe)))
        else:
            return len(node["children"]) - 1

    def connection(self):
        """draw connection"""
        pass

    def annotate(self):
        pass


class Hyperparam(Node):

    def __init__(self, hyperparam_name):
        self.name = hyperparam_name
        self.trainable = False

    def is_trainable_params(self):
        pass

    def annotate(self):
        pass



# ----------------------------------------------------------------------
# Created: fre mar 12 19:58:21 2021 (+0100)
# Last-Updated:
# Filename: ns_alex.py
# Author: Yinan Yu
# Description: alex namespace
# ----------------------------------------------------------------------
import inspect
import math
import collections.abc as collections
from collections import OrderedDict
from copy import deepcopy
import os
from pprint import pprint
from alex.alex import const, util
# ---------------------------------------------------------------------------- #


def indent(src: str, levels=1):
    return ("\t"*levels + src).expandtabs(4)


def wrap_in_function(src, fn, args, return_val_str=None):
    if not isinstance(args, list):
        args = [args]
    fn_src = "def %s(%s):\n" % (fn, ", ".join(args))
    if isinstance(src, str):
        src = src.split("\n")
    for line in src:
        fn_src += indent(line)
        if "\n" not in fn_src:
            fn_src += "\n"
    if return_val_str is None:
        return_val_str = line.split("=")[0]
    fn_src += indent("return %s\n" % return_val_str)
    return fn_src



# ----------------------------------------------------------------------
# Created: mån maj 17 21:52:56 2021 (+0200)
# Last-Updated:
# Filename: param_count.py
# Author: Yinan Yu
# Description:
# ----------------------------------------------------------------------
import numpy as np
from math import floor
from pprint import pprint
from copy import deepcopy
from alex.alex import core, const, node_interface
from alex.alex.annotator_interface import Annotator
from alex.engine import ns_alex


class Ingredient(node_interface.Ingredient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.required_input_rank = None
        self.required_input_num = None

    def get_shape(self, input_shape, *args):
        return input_shape

    def get_trainable_params_count(self, node, annotated):
        name = node["name"]
        return sum(list(map(lambda x: x["count"],
                            core.get_children(name,
                                              annotated))))

    def get_rank(self, node):
        return [len(node["meta"]["position"]["input_shape"]), # input rank
                len(node["shape"])] # output rank

    def get_required_input_rank(self):
        return self.required_input_rank


    def get_required_input_num(self):
        return self.required_input_num

    def validate_input_dims(self, node):
        # Validate if input dimensions are valid
        pass

    def validate_input_rank(self, node):
        required_input_rank = self.get_required_input_rank()
        if required_input_rank is not None:
            ranks = self.get_rank(node)
            if required_input_rank != ranks[0]:
                raise Exception("""Connection is invalid between
                %s (required rank %s) and %s (rank %s)""" % (node["name"],
                                                             str(required_input_rank),
                                                             str(node["meta"]["position"]["inputs"]),
                                                             str(ranks[0])))

    def validate_input_num(self, node):
        if self.get_required_input_num() is not None and node["meta"]["position"]["inputs"] is not None:
            if self.get_required_input_num() != len(node["meta"]["position"]["inputs"]):
                raise Exception("""Connection is invalud between
                %s (required inputs %s) and
                %s (inputs %s)""" % (node["name"],
                                     str(self.get_required_input_num()),
                                     str(node["meta"]["position"]["inputs"]),
                                     str(len(node["meta"]["position"]["inputs"]))))

    def validate_connection(self, node):
        self.validate_input_num(node)
        self.validate_input_rank(node)
        self.validate_input_dims(node)


class Recipe(node_interface.Recipe):

    def get_shape(self, input_shape, *args):
        return input_shape

    def get_trainable_params_count(self, node, annotated):
        name = node["name"]
        return sum(list(map(lambda x: x["count"],
                            core.get_children(name,
                                              annotated))))

    def validate_connection(self, node):
        pass


class Hyperparam(node_interface.Hyperparam):

    def get_trainable_params_count(self, node, annotated):
        if self.trainable:
            return np.prod(np.asarray(node["shape"]))
        else:
            return 0

    def get_shape(self, input_shape, *args):
        return [0, ]

    def validate_connection(self, node):
        pass

# ------------------------------------------------------------------------ #
class Dense(Ingredient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.required_input_rank = 1


class CrossEntropy(Ingredient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.required_input_num = 2


class Batchnorm(Ingredient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.required_input_rank = 3


class Add(Ingredient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.required_input_num = 2

    def validate_input_dims(self, node):
        # TODO
        pass


class Conv2D(Ingredient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_shape(self, input_shape, name, annotated):
        hyperparams = annotated[name]["meta"]["hyperparams"]
        strides = hyperparams["strides"]
        dilation = hyperparams["dilation"]
        padding = hyperparams["padding"]
        filters_shape = hyperparams["filters"]["shape"]
        k_h = int(filters_shape["kernel_size_h"])
        k_w = int(filters_shape["kernel_size_w"])
        n_filters = int(filters_shape["n_filters"])
        h = input_shape[0]
        w = input_shape[1]
        if padding == "SAME":
            padding = [1, 1] # FIXME
        elif padding == "VALID":
            padding = [0, 0]
        shape = [floor((h + 2*padding[0] - dilation*(k_h-1) - 1)/strides[0] + 1),
                 floor((w + 2*padding[1] - dilation*(k_w-1) - 1)/strides[1] + 1),
                 n_filters]
        return shape


class MaxPool2D(Ingredient):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.required_input_rank = 3

    def get_shape(self, input_shape, name, annotated):
        hyperparams = annotated[name]["meta"]["hyperparams"]
        padding = hyperparams["padding"]
        strides = hyperparams["strides"]
        dilation = hyperparams["dilation"]
        kernel_shape = hyperparams["window_shape"]

        k_h = int(kernel_shape[0])
        k_w = int(kernel_shape[1])
        h = input_shape[0]
        w = input_shape[1]
        if padding == "SAME":
            padding = [1, 1] # FIXME
        elif padding == "VALID":
            padding = [0, 0]
        shape = [floor((h + 2*padding[0] - dilation*(k_h-1) - 1)/strides[0] + 1),
                 floor((w + 2*padding[1] - dilation*(k_w-1) - 1)/strides[1] + 1),
                 input_shape[-1]]
        return shape


class Conv2DFilters(Hyperparam):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.trainable = True

    def get_shape(self, input_shape, name, annotated):
        shape = annotated[name]["meta"]["hyperparams"]["shape"]
        return [input_shape[-1],
                shape["kernel_size_h"],
                shape["kernel_size_w"],
                shape["n_filters"]]


class DenseWeights(Hyperparam):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.trainable = True

    def get_shape(self, input_shape, name, annotated):
        return [input_shape[-1],
                annotated[name]["meta"]["hyperparams"]["shape"]["n_units"]]


class DenseBias(Hyperparam):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.trainable = True

    def get_shape(self, input_shape, *args):
        return [1, ]


class BatchnormMean(Hyperparam):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_shape(self, input_shape, *args):
        return [input_shape[-1], ]


class BatchnormVariance(Hyperparam):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_shape(self, input_shape, *args):
        return [input_shape[-1], ]


class BatchnormOffset(Hyperparam):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.trainable = True

    def get_shape(self, input_shape, *args):
        return [input_shape[-1], ]


class BatchnormScale(Hyperparam):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.trainable = True

    def get_shape(self, input_shape, *args):
        return [input_shape[-1], ]


# None trainable
class Flatten(Ingredient):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.required_input_rank = None

    def get_shape(self, input_shape, *args):
        return [np.prod(np.asarray(input_shape)), ]


class Data(Ingredient):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_shape(self, shape):
        return [shape["height"], shape["width"], shape["channels"]]


class Label(Ingredient):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_shape(self, shape):
        return [shape["dim"], ]


def nodes(node):
    value = node["value"]
    nodes_ingredient = {"ingredient": Ingredient,
                        "data": Data,
                        "cross_entropy": CrossEntropy,
                        "label": Label,
                        "dense": Dense,
                        "add": Add,
                        "batch_normalize": Batchnorm,
                        "conv": Conv2D,
                        "max_pool2d": MaxPool2D,
                        "flatten": Flatten}
    nodes_hyperparam = {"hyperparam": Hyperparam}
    nodes_param = {"filters": Conv2DFilters,
                   "weights": DenseWeights,
                   "bias": DenseBias,
                   "means": BatchnormMean,
                   "vairance": BatchnormVariance,
                   "offset": BatchnormOffset,
                   "scale": BatchnormOffset}
    nodes_recipe = {"recipe": Recipe}

    _nodes = {**nodes_hyperparam,
              **nodes_ingredient,
              **nodes_param,
              **nodes_recipe}
    if value not in _nodes:
        value = node["type"]
    return _nodes[value](node["value"])


class ParamCount(Annotator):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.anno_name = "Trainable params count"
        self.passes = [[self.cache_input,
                        self.annotate_count]]

    def annotate_count(self, node):
        node = deepcopy(node)
        node["count"] = nodes(node).get_trainable_params_count(node=node,
                                                               annotated=self.annotated)
        return node

    def cache_input(self, node):
        """Cache the shape info in the node"""
        node = deepcopy(node)
        name = node["name"]
        input_nodes = self.get_input_nodes(name, self.annotated)
        if input_nodes is not None and input_nodes[0] is None:
            input_nodes = None
        if input_nodes is not None:
            # FIXME: do not handle validation for more than one inputs
            input_shape = input_nodes[0]["shape"]
            node["shape"] = nodes(node).get_shape(input_shape,
                                                  name,
                                                  self.annotated)
            if node["value"] == "channels":
                node["value"] = input_shape[-1]
                node["name"] = input_shape[-1]
        else:
            input_shape = None
            if name in self.components:
                node["type"] = "ingredient"
                component = self.components[name]
                hyperparams = component["value"]["hyperparams"]
                if "shape" in hyperparams:
                    node["shape"] = nodes(node).get_shape(hyperparams["shape"])
                else:
                    node["shape"] = None
        if input_nodes is None:
            inputs = None
        else:
            inputs = list(map(lambda x: x["name"], input_nodes))
        node["input_nodes"] = inputs
        node["meta"]["position"]["input_shape"] = input_shape
        node["ancestor"] = core.get_ancestor_ingredient_node(node, self.tree)
        if node["ancestor"]["name"] in self.components:
            node["dtype"] = self.components[node["ancestor"]["name"]]["meta"]["dtype"]
        else:
            node["dtype"] = None
        nodes(node).validate_connection(node)
        return node



# ----------------------------------------------------------------------
# Created: mån jun  3 15:56:29 2019 (+0200)
# Last-Updated:
# Filename: checkpoint.py
# Author: Yinan Yu
# Description:
# ----------------------------------------------------------------------
import json
import os
from copy import deepcopy
from pprint import pprint
import numpy as np
from pathlib import Path
from datetime import datetime
from alex.alex import const, util, dsl_parser, compare


def get_checkpoint(graph_list, states, trainable_params):
    network = dict()
    network["state"] = states
    network["components"] = []

    for component_block in graph_list:
        component = dict()
        component["value"] = dict()
        name = component_block["meta"]["name"]
        names_trainable_params = list(filter(lambda x: name+"/" in x,
                                             trainable_params))
        component["meta"] = deepcopy(component_block["meta"])
        for _key in component_block["value"]:
            if _key == const.TENSOR:
                component["value"][_key] = None
            elif _key == const.VAR:
                component["value"][_key] = dict()
                for _param in names_trainable_params:
                    # TODO: need to be framework dependent
                    # in both tensorflow and pytorch it happens to be .tolist()
                    component["value"][_key][_param] = trainable_params[_param].tolist()
            else:
                component["value"][_key] = deepcopy(component_block["value"][_key])
        network["components"].append(component)
    return network


def dump(graph_list, states, trainable_params, path):
    _network = get_checkpoint(graph_list, states, trainable_params)
    _json_cache = "/".join(path.split("/")[0:-1] + ["."+path.split("/")[-1]])
    with open(_json_cache, "w") as write_file:
        json.dump(_network, write_file, indent=4)
    os.rename(_json_cache, path)


def load_json(checkpoint_dir, json_file=None):
    json_file = get_load_path(checkpoint_dir, json_file)
    data = util.read_json(json_file)
    return data


def load(graph_list,
         checkpoint_dir,
         ckpt_name=None,
         matched=None):
    graph_list = deepcopy(graph_list)
    states = init_states()
    ckpt = load_json(checkpoint_dir, ckpt_name)
    trainable_params = dict()
    if ckpt is not None:
        states = ckpt["state"]
        ckpt_components = dsl_parser.list_to_dict(ckpt["components"])
        for i, component in enumerate(graph_list):
            name_in_new_config = component["meta"]["name"]
            if matched is None or name_in_new_config in matched:
                name_in_ckpt = matched[component["meta"]["name"]]
                _params = deepcopy(ckpt_components[name_in_ckpt]["value"]["var"])
                component["value"]["var"] = dict()
                for _param in _params:
                    _param_name = _param.split("/")[-1]
                    param_name = "%s/%s" % (name_in_new_config, _param_name)
                    trainable_params[param_name] = _params[_param]
                    graph_list[i]["value"]["var"][param_name] = _params[_param]

    return graph_list, trainable_params, states


############ Setup data:
def init_states(train_info={"meta": {"epoch": 0}},
                valid_info={"meta": dict(), "logs": {"loss": [],
                                                     "accuracy": []}},
                test_info=dict()):
    return {"valid": valid_info,
            "train": train_info,
            "test": test_info}


def save(graph_list, states, trainable_params, path):
    os.makedirs(path.split("/")[0], exist_ok=True)
    dump(graph_list, states, trainable_params, path)


def get_checkpoint_path(ckpt_dir, ckpt_name=None, checkpoint_ext=".json"):
    if ckpt_name is None:
        _id = "_" + str(datetime.timestamp(datetime.now())).replace(".", "")
        ckpt_name = "config" + _id
    if checkpoint_ext not in ckpt_name:
        ckpt_name += checkpoint_ext
    return os.path.join(ckpt_dir,
                        ckpt_name)


def get_load_path(ckpt_dir, ckpt_name=None, ext=".json"):
    json_file = None
    if ckpt_name is not None:
        if ckpt_name.split(".")[0] == "latest":
            json_file = util.get_latest_file(ckpt_dir,
                                             ext=ext)
        else:
            if ext not in ckpt_name:
                ckpt_name += ext
            json_file = os.path.join(ckpt_dir,
                                     ckpt_name)
    if json_file is None or not Path(json_file).is_file():
        print("Checkpoint %s does not exist" % json_file)
        return None
    else:
        return json_file


class Checkpoint():

    def __init__(self,
                 config,
                 load=["log", None], # [dir, path]
                 save=["log", None]): # [dir, path]

        self.components_list = dsl_parser.parse(config)
        self.load_path = get_load_path(ckpt_dir=load[0],
                                       ckpt_name=load[1])
        self.save_path = get_checkpoint_path(ckpt_dir=save[0],
                                             ckpt_name=save[1])
        self.save_dir = "/".join(self.save_path.split("/")[:-1])
        self.save_name = self.save_path.split("/")[-1]

        if self.load_path is not None:
            self.load_dir = "/".join(self.load_path.split("/")[:-1])
            self.load_name = self.load_path.split("/")[-1]
            # Check the diff between config and load:
            loaded_component_list = load_json(self.load_dir,
                                              self.load_name)["components"]
            self.matched = compare.matched_ingredients(self.components_list,
                                                       loaded_component_list,
                                                       os.path.join(self.save_dir,
                                                                    "ckpt_diff.png"))
        else:
            self.matched = None

    def load(self):
        if self.load_path is None:
            trainable_params = None
            stats = None
        else:
            self.components_list, trainable_params, stats = load(self.components_list,
                                                                 self.load_dir,
                                                                 ckpt_name=self.load_name,
                                                                 matched=self.matched)
        return trainable_params

    def save(self, trainable_params=dict(), stats={}):
        save(self.components_list, stats, trainable_params, self.save_path)
        print("Saving to path", self.save_path)



# ----------------------------------------------------------------------
# Created: mån maj 31 18:55:54 2021 (+0200)
# Last-Updated:
# Filename: test_checkpoint.py
# Author: Yinan
# Description:
# ----------------------------------------------------------------------

import unittest
from alex.alex import checkpoint
from pprint import pprint


class TestCheckpoint(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        print("Testing checkpoint")

    def setUp(self):
        pass

    def test_checkpoint(self):
        ckpt = checkpoint.Checkpoint("examples/configs/small2.yml",
                                     ["checkpoints",
                                      "config_1625826546926544.json"],
                                     ["checkpoints", None])
        # ckpt.save()
        print("Matched components")
        pprint(ckpt.matched)


if __name__ == '__main__':
    unittest.main()



# ----------------------------------------------------------------------
# Created: tor mar 25 10:03:41 2021 (+0100)
# Last-Updated:
# Filename: test_util.py
# Author: Yinan Yu
# Description:
# ----------------------------------------------------------------------

from alex.alex import util
import unittest
from collections import OrderedDict, Iterable


class TestUtil(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        pass

    def setUp(self):
        pass

    def test_fill_dict(self):
        dict1 = {"fn1": {}, "fn2": {}}
        dict1["fn1"]["var1"] = "def1"
        dict1["fn1"]["var2"] = "def2"
        dict1["fn2"]["var1"] = "def1"
        dict2 = {"fn2": {"var2": "def3"}}
        util.fill_dict(dict1, dict2)
        self.assertDictEqual(dict2, {"fn1": {"var1": "def1",
                                             "var2": "def2"},
                                     "fn2": {"var1": "def1",
                                             "var2": "def3"}})

    def test_updateordered_dict(self):
        def _get_dict1():
            dict1 = {"fn1": OrderedDict(),
                     "fn2": OrderedDict()}
            dict1["fn1"]["var1"] = "def1"
            dict1["fn1"]["var2"] = "def2"
            dict1["fn2"]["var1"] = "def1"
            dict1["fn2"]["var2"] = "def2"
            return dict1

        dict1 = _get_dict1()
        dict2 = {"fn2": {"var2": "def3"}}
        util.update_ordereddict(dict2, dict1)
        self.assertDictEqual(dict1, OrderedDict({"fn1": {"var1": "def1",
                                                         "var2": "def2"},
                                                 "fn2": {"var1": "def1",
                                                         "var2": "def3"}}))
        dict1 = _get_dict1()
        dict2 = {"fn3": {"var2": "def3"}}
        util.update_ordereddict(dict2, dict1)
        self.assertDictEqual(dict1, OrderedDict({"fn1": {"var1": "def1",
                                                         "var2": "def2"},
                                                 "fn2": {"var1": "def1",
                                                         "var2": "def2"},
                                                 "fn3": {"var2": "def3"}}))

        dict1 = _get_dict1()
        dict2 = {"fn2": 0}
        util.update_ordereddict(dict2, dict1)
        self.assertDictEqual(dict1, OrderedDict({"fn1": {"var1": "def1",
                                                         "var2": "def2"},
                                                 "fn2": 0}))

        dict1 = _get_dict1()
        dict1["fn3"] = {"var1": [1, 2, 3]}
        dict2 = {"fn3": {"var1": [2, 3, 4]}}
        util.update_ordereddict(dict2, dict1)
        self.assertDictEqual(dict1, OrderedDict({"fn1": {"var1": "def1",
                                                         "var2": "def2"},
                                                 "fn2": {"var1": "def1",
                                                         "var2": "def2"},
                                                 "fn3": {"var1": [1, 2, 3, 4]}}))


        dict1 = _get_dict1()
        dict1["fn3"] = {"var1": [1, 2, 3], "var2": "def2"}
        dict2 = {"fn3": {"var1": [2, 3, 4], "var2": "def5"}}
        util.update_ordereddict(dict2, dict1)
        self.assertDictEqual(dict1, OrderedDict({"fn1": {"var1": "def1",
                                                         "var2": "def2"},
                                                 "fn2": {"var1": "def1",
                                                         "var2": "def2"},
                                                 "fn3": {"var1": [1, 2, 3, 4],
                                                         "var2": "def5"}}))

    def test_get_value(self):

        dict1 = {"a": 1,
                 "b": {"c": {"d": 4},
                       "e": {"f": {"g": 2}, "h": 5}},
                 "i": {"j": 100}}
        res, found = util.get_value(dict1, "g", found=False)
        self.assertEqual(res, 2)
        self.assertEqual(found, True)
        res, found = util.get_value(dict1, "i", found=False)
        self.assertDictEqual(res, {"j": 100})
        self.assertEqual(found, True)

    def test_replace_key(self):
        dict1 = {"a": 1,
                 "b": {"c": {"d": 4},
                       "e": {"f": {"g": 2}, "h": 5}},
                 "i": {"j": 100}}
        dict2 = {"a": 1,
                 "b": {"c": {"d": 4},
                       "e": {"f": 99, "h": 5}},
                 "i": {"j": 100}}
        util.replace_key(dict1, "f", 99)
        self.assertDictEqual(dict1, dict2)
        dict1 = {"a": 1,
                 "b": {"c": {"d": 4},
                       "e": {"f": {"g": 2}, "h": 5}},
                 "i": {"j": 100}}
        dict3 = {"a": 1,
                 "b": {"c": {"d": 4},
                       "e": {"f": {"k": {"l": 1}}, "h": 5}},
                 "i": {"j": 100}}
        util.replace_key(dict1, "f", {"k": {"l": 1}})
        self.assertDictEqual(dict1, dict3)

    def test_flatten_dict_keys(self):
        dict1 = {"a": 1,
                 "b": {"c": {"d": 4},
                       "e": {"f": {"g": 2}, "h": 5}}}
        ancestors, key_lst = util.flatten_dict_keys(dict1)
        self.assertSetEqual(set(key_lst),
                            {"a", "b",
                             "b/c/d", "b/c",
                             "b/e", "b/e/f", "b/e/f/g", "b/e/h"})

    def test_flatten_subdict_keys(self):
        dict1 = {"a": 1,
                 "b": {"c": {"d": 4},
                       "e": {"f": {"g": 2}, "h": 5}}}
        ancestors, key_lst = util.flatten_subdict_keys(dict1)
        self.assertSetEqual(set(key_lst),
                            {"a", "b",
                             "c/d", "b/c",
                             "b/e", "e/f", "f/g", "e/h"})

    def test_keys_unique(self):
        dict1 = {"a": 1,
                 "b": {"c": {"d": 4},
                       "e": {"f": {"g": 2}, "h": 5}}}
        self.assertTrue(util.keys_unique(dict1))
        dict2 = {"a": 1,
                 "b": {"c": {"d": 4},
                       "b": {"f": {"g": 2}, "h": 5}}}

        self.assertTrue(not util.keys_unique(dict2))


if __name__ == '__main__':
    unittest.main()



# ----------------------------------------------------------------------
# Created: mån mar 29 11:10:21 2021 (+0200)
# Last-Updated:
# Filename: ns_pytorch.py
# Author: Yinan Yu
# Description:
# ----------------------------------------------------------------------
from alex.alex import const
from alex.engine import ns_alex


ENGINE = "pytorch"
DIM_ORDER = ["batch_size", "channel", "height", "width"]


def add_imports(additional_modules=[]):
    default_modules = [["torch"],
                       ["numpy", "np"]]
    modules = default_modules + additional_modules
    imports = ""
    for m in modules:
        imports += "import %s " % m[0]
        if len(m) == 2:
            imports += "as %s\n" % m[1]
        elif len(m) == 1:
            imports += "\n"
    imports += "\n\n"
    return imports


def add_global_configs():
    deterministic = "torch.backends.cudnn.deterministic = True\n"
    device = "device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')\n"
    dtypes = "torch_types = {'float32': torch.float32, 'int8': torch.int8}\n"
    space = "\n\n"
    return deterministic + device + dtypes + space


def wrap_in_class(trainable_params_code, component_code, loss_code, optimizer_code, scheduler_code=""):
    code = []
    code.append("class Model(torch.nn.Module):")
    code.append("")
    code.append(ns_alex.indent("def __init__(self, ckpt=None):", levels=1))
    code.append(ns_alex.indent("super(Model, self).__init__()", levels=2))
    code.append(ns_alex.indent("self.%s = self.get_trainable_params(ckpt)" % (const.ALEX_ARG_TRAINABLE_PARAMS), levels=2))
    code.append(ns_alex.indent("self.params = []", levels=2))
    code.append(ns_alex.indent("for var in self.%s:" % const.ALEX_ARG_TRAINABLE_PARAMS, levels=2))
    code.append(ns_alex.indent("self.register_parameter(var, self.%s[var])" % const.ALEX_ARG_TRAINABLE_PARAMS, levels=3))
    code.append(ns_alex.indent("self.params.append({'params': self.%s[var]})" % const.ALEX_ARG_TRAINABLE_PARAMS, levels=3))
    code.append("")
    code.append(ns_alex.indent("def forward(self, x, training):", levels=1))
    code.append(ns_alex.indent("x = self.model(x, self.%s, training)" % const.ALEX_ARG_TRAINABLE_PARAMS,
                               levels=2))
    code.append(ns_alex.indent("return x", levels=2))
    code.append("")
    code.append(ns_alex.indent("@staticmethod", levels=1))
    code.append(ns_alex.indent(trainable_params_code.replace("\n", "\n\t"), levels=1))
    code.append(ns_alex.indent("@staticmethod", levels=1))
    code.append(ns_alex.indent(component_code.replace("\n", "\n\t"), levels=1))
    code.append(ns_alex.indent("@staticmethod", levels=1))
    code.append(ns_alex.indent(loss_code.replace("\n", "\n\t"), levels=1))
    code.append(ns_alex.indent("@staticmethod", levels=1))
    code.append(ns_alex.indent(optimizer_code.replace("\n", "\n\t"), levels=1))
    if scheduler_code != "":
        code.append(ns_alex.indent("@staticmethod", levels=1))
        code.append(ns_alex.indent(scheduler_code.replace("\n", "\n\t"), levels=1))
    code = "\n".join(code)
    return code



# ----------------------------------------------------------------------
# Created: tis feb 23 22:32:09 2021 (+0100)
# Last-Updated:
# Filename: setup.py
# Author: Yinan Yu
# Description:
# ----------------------------------------------------------------------

import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="alex-nn", # Replace with your own username
    version="0.0.9",
    author="Yinan Yu",
    author_email="yu.yinan16@gmail.com",
    description="Network analysis",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/yinan16/alex",
    project_urls={
        "Bug Tracker": "https://github.com/yinan16/alex/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
    ],
    install_requires=[
        "pandas",
        "numpy",
        "pydot==1.4.1",
        "pyparsing",
        "scikit-learn==0.23.2",
        "PyYAML>=5.4",
        "jsonschema",
        "rope",
        "matplotlib",
    ],
    scripts=["bin/alex-nn"],
    include_package_data=True,
    package_dir={"": "./"},
    packages=["alex",
              "alex.components",
              "alex.annotators",
              "alex.engine",
              "alex.alex"],
    package_data={'alex.components': ['*.yml']},
    python_requires=">=3.6",
)
